class ComponentIncompleteError(Exception):
    pass


class DimensionNotSuitableForEnv(Exception):
    pass


class TrajectoryNotSupported(Exception):
    pass


class TrajectoryComponentMissingError(Exception):
    pass


class JointSpaceGoalsNotSupportedError(Exception):
    pass


class MultiplePrimeGoalsError(Exception):
    pass


class UnknownSubGoalType(Exception):
    pass


class MissmatchDimensionError(Exception):
    pass


class NoDistanceFunctionForURDFObstacle(Exception):
    pass



import numpy as np
import sympy as sp

from mpscenes.common.reference_trajectory import ReferenceTrajectory, t


def evaluate_component(expression, evaluation_time: float):
    evaluation = []
    for expr in expression:
        if isinstance(expr, float):
            evaluation.append(expr)
        else:
            evaluation.append(float(expr.evalf(subs={"t": evaluation_time})))
    return np.array(evaluation)


class AnalyticTrajectory(ReferenceTrajectory):
    def __init__(self, n: int, **kwargs):
        super().__init__(n, **kwargs)
        self._t = t
        self._traj = [eval(x) for x in self.trajectory_dictionary()]

    def concretize(self):
        self._velocity = [sp.diff(component, t) for component in self._traj]
        self._acceleration = [sp.diff(component, t) for component in self._velocity]

    def shuffle(self):
        pass

    def evaluate(self, time_step: float) -> list:
        position = evaluate_component(self._traj, time_step)
        velocity = evaluate_component(self._velocity, time_step)
        acceleration = evaluate_component(self._acceleration, time_step)
        return [position, velocity, acceleration]



from abc import ABC
import yaml
from omegaconf import OmegaConf

from mpscenes.common.errors import (
    ComponentIncompleteError,
    DimensionNotSuitableForEnv,
)


class MPComponent(ABC):
    def __init__(self, **kwargs):
        schema = kwargs.get("schema")
        if "content_dict" in kwargs and "name" in kwargs:
            self._content_dict = kwargs.get("content_dict")
            self._name = kwargs.get("name")
        elif "file_name" in kwargs:
            with open(kwargs.get("file_name"), "r") as stream:
                self._content_dict = yaml.safe_load(stream)
            self._name = self._content_dict["name"]
            del self._content_dict["name"]
        self._config = OmegaConf.create(self._content_dict)
        config = OmegaConf.create(self._content_dict)
        self._config = OmegaConf.merge(schema, config)

    def check_completeness(self):
        pass
        """
        incomplete = False
        missingKeys = ""
        for key in self._required_keys:
            if key not in self._content_dict.keys():
                incomplete = True
                missingKeys += key + ", "
        if incomplete:
            raise ComponentIncompleteError("Missing keys: %s" % missingKeys[:-2])
        """

    def name(self):
        return self._name

    def evaluate_components(self, mask: list, t: float):
        result = {}
        for mask_entry in mask:
            try:
                result[mask_entry] = getattr(self, mask_entry)(t=t)
            except TypeError as _:
                result[mask_entry] = getattr(self, mask_entry)()
        return result

    def dict(self):
        return OmegaConf.to_container(self._config)



from dataclasses import dataclass
from typing import List, Optional, Any
import numpy as np
from omegaconf import OmegaConf
from pyquaternion import Quaternion
from mpscenes.goals.sub_goal import SubGoal, SubGoalConfig
from mpscenes.common.errors import DimensionNotSuitableForEnv, TrajectoryNotSupported
from mpscenes.common.analytic_trajectory import AnalyticTrajectory
from mpscenes.common.spline_trajectory import SplineTrajectory


@dataclass
class DynamicSubGoalConfig(SubGoalConfig):
    """Configuration dataclass for static sub goal.

    This configuration class holds information about the
    the weight, accuracy required, type and position in the
    kinematic chain.

    Parameters:
    ------------

    parent_link: str
        Name of the link that specifies the frame in which the goal is defined
    child_link: str
        Name of the link that should match the desired position
    trajectory: Any
        Trajectory of the goal, either defined by spline or analytic equation.
    angle list
        Additional rotation from the parent_link frame given by a quaternion
    low : list
        Lower limit for randomization
    high : list
        Upper limit for randomization

    """

    parent_link: str
    child_link: str
    trajectory: Any
    angle: Optional[Any] = None
    low: Optional[List[float]] = None
    high: Optional[List[float]] = None


class DynamicSubGoal(SubGoal):
    def __init__(self, **kwargs):
        if not 'schema' in kwargs:
            schema = OmegaConf.structured(DynamicSubGoalConfig)
            kwargs['schema'] = schema
        super().__init__(**kwargs)
        if "controlPoints" in self._config.trajectory:
            self._trajectory_type = "spline"
        elif isinstance(self._config.trajectory[0], str):
            self._trajectory_type = "analytic"
        else:
            raise TrajectoryNotSupported(
                f"Trajectory definition not supported, {self._config.trajectory}"
            )
        self.check_completeness()
        if self.trajectory_type() == 'spline':
            self._traj = SplineTrajectory(
                self.dimension(), traj=self._config.trajectory
            )
        elif self.trajectory_type() == "analytic":
            self._traj = AnalyticTrajectory(
                self.dimension(), traj=self._config.trajectory
            )
        self._traj.concretize()
        self.check_dimensionality()

    def trajectory_type(self) -> str:
        return self._trajectory_type

    def parent_link(self):
        return self._config.parent_link

    def child_link(self):
        return self._config.child_link

    def traj(self):
        return self._traj

    def dict(self):
        return OmegaConf.to_container(self._config)

    def position(self, **kwargs) -> np.ndarray:
        return self.evaluate_trajectory(**kwargs)[0]

    def velocity(self, **kwargs) -> np.ndarray:
        return self.evaluate_trajectory(**kwargs)[1]

    def acceleration(self, **kwargs) -> np.ndarray:
        return self.evaluate_trajectory(**kwargs)[2]

    def evaluate_trajectory(self, **kwargs):
        t = kwargs.get("t") if "t" in kwargs else 0.0
        return self._traj.evaluate(time_step=t)

    def shuffle(self):
        self._traj.shuffle()

    def angle(self):
        return self._config.angle



from dataclasses import dataclass
from typing import Optional
from omegaconf import OmegaConf

from mpscenes.obstacles.dynamic_obstacle import DynamicObstacle, DynamicGeometryConfig
from mpscenes.obstacles.box_obstacle import BoxGeometryConfig, BoxObstacle, BoxObstacleConfig


@dataclass
class DynamicBoxGeometryConfig(BoxGeometryConfig, DynamicGeometryConfig):
    pass

@dataclass
class DynamicBoxObstacleConfig(BoxObstacleConfig):
    """Configuration dataclass for sphere obstacle.

    This configuration class holds information about the position, size
    and randomization of a dynamic spherical obstacle.

    Parameters:
    ------------

    geometry : DynamicBoxGeometryConfig : Geometry of the obstacle
    low: DynamicBoxGeometryConfig : Lower limit for randomization
    high: DynamicBoxGeometryConfig : Upper limit for randomization
    """

    geometry: DynamicBoxGeometryConfig
    low: Optional[DynamicBoxGeometryConfig] = None
    high: Optional[DynamicBoxGeometryConfig] = None


class DynamicBoxObstacle(DynamicObstacle, BoxObstacle):

    def __init__(self, **kwargs):
        schema = OmegaConf.structured(DynamicBoxObstacleConfig)
        kwargs['schema'] = schema
        super().__init__(**kwargs)



from dataclasses import dataclass
from typing import List, Optional
import numpy as np
from omegaconf import OmegaConf

from mpscenes.obstacles.collision_obstacle import CollisionObstacle, CollisionObstacleConfig, GeometryConfig
from mpscenes.common.errors import NoDistanceFunctionForURDFObstacle, DimensionNotSuitableForEnv


@dataclass
class UrdfGeometryConfig(GeometryConfig):
    pass


@dataclass
class UrdfObstacleConfig:
    """Configuration dataclass for sphere obstacle.

    This configuration class holds information about the position, size
    and randomization of a spherical obstacle.

    Parameters:
    ------------

    urdf : str : Filename of the urdf
    geometry : GeometryConfig : Geometry of the obstacle
    low: GeometryConfig : Lower limit for randomization
    high: GeometryConfig : Upper limit for randomization

    """

    urdf: str
    type: str
    geometry: UrdfGeometryConfig
    scaling: float = 1
    movable: bool = False
    low: Optional[UrdfGeometryConfig] = None
    high: Optional[UrdfGeometryConfig] = None


class UrdfObstacle(CollisionObstacle):
    def __init__(self, **kwargs):
        if not 'schema' in kwargs:
            schema = OmegaConf.structured(UrdfObstacleConfig)
            kwargs['schema'] = schema
        super().__init__(**kwargs)
        self.check_completeness()

    def size(self):
        return []

    def urdf(self):
        return self._config.urdf

    def position(self):
        return self._config.geometry.position

    def velocity(self):
        return np.zeros(3)

    def acceleration(self):
        return np.zeros(3)

    def dimension(self):
        return len(self._config.geometry.position)

    def distance(self, position: np.ndarray) -> None:
        raise NoDistanceFunctionForURDFObstacle("Cannot compute distance for urdf-obstacle.")




import numpy as np

from geomdl import BSpline
from geomdl import utilities
import logging

from mpscenes.common.errors import TrajectoryComponentMissingError
from mpscenes.common.reference_trajectory import ReferenceTrajectory


class SplineTrajectory(ReferenceTrajectory):
    def __init__(self, n: int, **kwargs):
        super().__init__(n)
        self._config = kwargs
        if "degree" not in self.trajectory_dictionary().keys():
            raise TrajectoryComponentMissingError(
                "Spline definition not complete. Missing component in trajectory: degree"
            )
        if "controlPoints" not in self.trajectory_dictionary().keys():
            raise TrajectoryComponentMissingError(
                "Spline definition not complete. Missing component in trajectory: controlPoints"
            )
        if "duration" not in self.trajectory_dictionary().keys():
            raise TrajectoryComponentMissingError(
                "Spline definition not complete. Missing component in trajectory: duration"
            )
        self.initialize_spline()



    def initialize_spline(self):
        self._traj = BSpline.Curve()
        self._traj.degree = self.trajectory_dictionary()["degree"]
        list_ctrlpts = [list(val) for val in self.trajectory_dictionary()["controlPoints"]]
        self._traj.ctrlpts = list_ctrlpts
        self._traj.knotvector = utilities.generate_knot_vector(
            self._traj.degree, len(self._traj.ctrlpts)
        )
        self._duration = self.trajectory_dictionary()["duration"]

    def concretize(self):
        pass

    def shuffle(self):
        limit_low = np.array(self.trajectory_dictionary()["low"]["controlPoints"])
        limit_high = np.array(self.trajectory_dictionary()["high"]["controlPoints"])
        self._config['traj']["controlPoints"] = np.random.uniform(
            limit_low, limit_high, limit_low.shape
        ).tolist()
        self.initialize_spline()

    def timeReparameterize(self, time_step: float) -> float:
        if time_step > self._duration:
            return 1.0
        return 0.5 * (1 - np.cos(np.pi * time_step / self._duration))

    def getScaledDerivatives(self, time_step: float) -> list:
        t_ref = self.timeReparameterize(time_step)
        xds = self._traj.derivatives(t_ref, order=2)
        position = np.array(xds[0])
        v_raw = np.array(xds[1])
        a_raw = np.array(xds[2])
        v_scaling = (
            1 * np.pi / self._duration * np.sin(time_step * np.pi / self._duration)
        )
        a_scaling = (
            1 * np.pi / self._duration * np.cos(time_step * np.pi / self._duration)
        )
        velocity = v_scaling * v_raw / np.linalg.norm(v_raw)
        if np.linalg.norm(a_raw) < 1e-5:
            logging.warn(
                f"Assuming zero acceleration on spline, because of large magnitude {np.linalg.norm(a_raw)}"
            )
            acceleration = a_raw * 0
        else:
            acceleration = a_scaling * a_raw / np.linalg.norm(a_raw)
        if t_ref == 1.0:
            velocity = v_raw * 0.0
            acceleration = a_raw * 0.0
        return [position, velocity, acceleration]

    def evaluate(self, time_step: float) -> list:
        return self.getScaledDerivatives(time_step)



from dataclasses import dataclass
from omegaconf import OmegaConf
from typing import Optional

from mpscenes.obstacles.dynamic_obstacle import DynamicObstacle, DynamicGeometryConfig
from mpscenes.obstacles.sphere_obstacle import SphereGeometryConfig, SphereObstacle, SphereObstacleConfig


@dataclass
class DynamicSphereGeometryConfig(SphereGeometryConfig, DynamicGeometryConfig):
    pass

@dataclass
class DynamicSphereObstacleConfig(SphereObstacleConfig):
    """Configuration dataclass for sphere obstacle.

    This configuration class holds information about the position, size
    and randomization of a dynamic spherical obstacle.

    Parameters:
    ------------

    geometry : DynamicSphereGeometryConfig : Geometry of the obstacle
    low: DynamicSphereGeometryConfig : Lower limit for randomization
    high: DynamicSphereGeometryConfig : Upper limit for randomization
    """

    geometry: DynamicSphereGeometryConfig
    low: Optional[DynamicSphereGeometryConfig] = None
    high: Optional[DynamicSphereGeometryConfig] = None


class DynamicSphereObstacle(DynamicObstacle, SphereObstacle):

    def __init__(self, **kwargs):
        schema = OmegaConf.structured(DynamicSphereObstacleConfig)
        kwargs['schema'] = schema
        super().__init__(**kwargs)



from abc import ABC, abstractmethod
import sympy as sp

from mpscenes.common.errors import TrajectoryComponentMissingError

t = sp.symbols("t")


class ReferenceTrajectory(ABC):
    def __init__(self, n: int, **kwargs):
        assert isinstance(n, int)
        self._n = n
        self._config = kwargs

    def n(self):
        return self._n

    def trajectory_dictionary(self) -> dict:
        if "traj" not in self._config:
            raise TrajectoryComponentMissingError(
                "Reference Trajectory definition not complete. Missing component: traj"
            )
        return self._config["traj"]

    @abstractmethod
    def concretize(self):
        pass

    @abstractmethod
    def evaluate(self, time_step: float):
        pass



from dataclasses import dataclass
from typing import Optional
import numpy as np
from omegaconf import OmegaConf

from mpscenes.obstacles.collision_obstacle import CollisionObstacle, CollisionObstacleConfig, GeometryConfig


@dataclass
class BoxGeometryConfig(GeometryConfig):
    """Configuration dataclass for geometry.

    This configuration class holds information about position
    and size of the box obstacle.

    Parameters:
    ------------

    length: float: Length of the box
    width: float: Width of the box
    height: float: Height of the box
    """

    length: float = 1.0
    width: float = 1.0
    height: float = 1.0


@dataclass
class BoxObstacleConfig(CollisionObstacleConfig):
    """Configuration dataclass for box obstacle.

    This configuration class holds information about the position, size
    and randomization of a rectengular obstacle.

    Parameters:
    ------------

    geometry : BoxGeometryConfig : Geometry of the box
    low : BoxGeometryConfig : Lower limit for randomization
    high : BoxGeometryConfig : Upper limit for randomization
    """

    geometry: BoxGeometryConfig
    low: Optional[BoxGeometryConfig] = None
    high: Optional[BoxGeometryConfig] = None


class BoxObstacle(CollisionObstacle):
    def __init__(self, **kwargs):
        if not 'schema' in kwargs:
            schema = OmegaConf.structured(BoxObstacleConfig)
            kwargs['schema'] = schema
        super().__init__(**kwargs)
        self.check_completeness()

    def size(self):
        return [
            self.length(),
            self.width(),
            self.height(),
        ]

    def dimension(self):
        return len(self._config.geometry.position)

    def limit_low(self):
        if self._config.low:
            return [
                np.array(self._config.low.position),
                self._config.low.length,
                self._config.low.width,
                self._config.low.height,
            ]
        else:
            return [np.ones(self.dimension()) * -1, 0, 0, 0]

    def limit_high(self):
        if self._config.high:
            return [
                np.array(self._config.high.position),
                self._config.high.length,
                self._config.high.width,
                self._config.high.height,
            ]
        else:
            return [np.ones(self.dimension()) * 1, 1, 1, 1]

    def position(self, **kwargs) -> np.ndarray:
        return np.array(self._config.geometry.position)

    def velocity(self, **kwargs) -> np.ndarray:
        return np.zeros(self.dimension())

    def acceleration(self, **kwargs) -> np.ndarray:
        return np.zeros(self.dimension())

    def length(self):
        return self._config.geometry.length

    def width(self):
        return self._config.geometry.width

    def height(self):
        return self._config.geometry.height

    def shuffle(self):
        random_pos = np.random.uniform(
            self.limit_low()[0], self.limit_high()[0], self.dimension()
        )
        random_length = np.random.uniform(
            self.limit_low()[1], self.limit_high()[1], 1
        )
        random_width = np.random.uniform(
            self.limit_low()[2], self.limit_high()[2], 1
        )
        random_height = np.random.uniform(
            self.limit_low()[3], self.limit_high()[3], 1
        )
        self._config.geometry.position = random_pos.tolist()
        self._config.geometry.length = float(random_length)
        self._config.geometry.width = float(random_width)
        self._config.geometry.height = float(random_height)

    def movable(self):
        return self._config.movable

    def csv(self, file_name, samples=100):
        pass

    def distance(self, position: np.ndarray, **kwargs) -> float:
        pos = self.position_into_obstacle_frame(position, **kwargs)
        q = np.transpose(np.subtract(np.transpose(np.absolute(pos)),
                                     np.array(self.size())/2.0)) 
        return np.linalg.norm(np.maximum(q, 0), axis=0) + np.minimum(np.maximum(q[0], np.maximum(q[1], q[2])), 0.0)



import yaml
from omegaconf import OmegaConf

from mpscenes.common.errors import MultiplePrimeGoalsError
from mpscenes.common.component import MPComponent
from mpscenes.goals.sub_goal import SubGoal
from mpscenes.goals.sub_goal_creator import SubGoalCreator


class GoalComposition(MPComponent):
    def __init__(self, **kwargs):
        if "content_dict" in kwargs and "name" in kwargs:
            self._content_dict = kwargs.get("content_dict")
            self._name = kwargs.get("name")
        elif "file_name" in kwargs:
            with open(kwargs.get("file_name"), "r") as stream:
                self._content_dict = yaml.safe_load(stream)
            self._name = self._content_dict["name"]
            del self._content_dict["name"]
        self._config = OmegaConf.create(self._content_dict)
        self._primary_goal_index = -1
        self._sub_goals = []
        self._sub_goal_creator = SubGoalCreator()
        self.parse_sub_goals()

    def parse_sub_goals(self):
        for sub_goal_name in self._config.keys():
            sub_goal_type = self._config[sub_goal_name].type
            sub_goal_dict = self._config[sub_goal_name]
            sub_goal = self._sub_goal_creator.create_sub_goal(
                sub_goal_type, sub_goal_name, sub_goal_dict
            )
            if sub_goal.is_primary_goal():
                if self._primary_goal_index >= 0:
                    raise MultiplePrimeGoalsError(
                        "There are multiple prime goals. Using first prime goal"
                    )
                else:
                    self._primary_goal_index = len(self._sub_goals)
            self._sub_goals.append(sub_goal)

    def primary_goal(self):
        return self.get_goal_by_index(self._primary_goal_index)

    def sub_goals(self):
        return self._sub_goals

    def get_goal_by_name(self, name) -> SubGoal:
        for sub_goal in self._sub_goals:
            if sub_goal.name() == name:
                return sub_goal

    def get_goal_by_index(self, index):
        return self._sub_goals[index]

    def evaluate(self, **kwargs):
        evals = []
        for sub_goal in self._sub_goals:
            evals += sub_goal.evaluate(**kwargs)
        return evals

    def dict(self):
        composition_dict = {}
        for sub_goal in self._sub_goals:
            composition_dict[sub_goal.name()] = sub_goal.dict()
        return composition_dict

    def shuffle(self):
        for sub_goal in self._sub_goals:
            sub_goal.shuffle()




from dataclasses import dataclass
from typing import Optional
from omegaconf import OmegaConf

from mpscenes.obstacles.dynamic_obstacle import DynamicObstacle, DynamicGeometryConfig
from mpscenes.obstacles.cylinder_obstacle import CylinderGeometryConfig, CylinderObstacle, CylinderObstacleConfig


@dataclass
class DynamicCylinderGeometryConfig(CylinderGeometryConfig, DynamicGeometryConfig):
    pass

@dataclass
class DynamicCylinderObstacleConfig(CylinderObstacleConfig):
    """Configuration dataclass for sphere obstacle.

    This configuration class holds information about the position, size
    and randomization of a dynamic spherical obstacle.

    Parameters:
    ------------

    geometry : DynamicCylinderGeometryConfig : Geometry of the obstacle
    low: DynamicCylinderGeometryConfig : Lower limit for randomization
    high: DynamicCylinderGeometryConfig : Upper limit for randomization
    """

    geometry: DynamicCylinderGeometryConfig
    low: Optional[DynamicCylinderGeometryConfig] = None
    high: Optional[DynamicCylinderGeometryConfig] = None


class DynamicCylinderObstacle(DynamicObstacle, CylinderObstacle):

    def __init__(self, **kwargs):
        schema = OmegaConf.structured(DynamicCylinderObstacleConfig)
        kwargs['schema'] = schema
        super().__init__(**kwargs)



from typing import List, Optional, Any
from dataclasses import dataclass
import numpy as np
from pyquaternion import Quaternion
from omegaconf import OmegaConf

from mpscenes.goals.sub_goal import SubGoal, SubGoalConfig
from mpscenes.common.errors import DimensionNotSuitableForEnv


@dataclass
class StaticSubGoalConfig(SubGoalConfig):
    """Configuration dataclass for static sub goal.

    This configuration class holds information about the
    the weight, accuracy required, type and position in the
    kinematic chain.

    Parameters:
    ------------

    parent_link: str
        Name of the link that specifies the frame in which the goal is defined
    child_link: str
        Name of the link that should match the desired position
    desired_position : list
        Goal state of the concerned link
    angle: list
        Additional rotation from the parent_link frame given by a quaternion
    low: list
        Lower limit for randomization
    high: list
        Upper limit for randomization

    """

    parent_link: Any
    child_link: Any
    desired_position: List[float]
    angle: Optional[Any] = None
    low: Optional[List[float]] = None
    high: Optional[List[float]] = None


class StaticSubGoal(SubGoal):
    def __init__(self, **kwargs):
        if not 'schema' in kwargs:
            schema = OmegaConf.structured(StaticSubGoalConfig)
            kwargs['schema'] = schema
        super().__init__(**kwargs)
        self.check_completeness()
        self.check_dimensionality()

    def parent_link(self):
        return self._config.parent_link

    def child_link(self):
        return self._config.child_link

    def limit_low(self):
        if self._config.low:
            return np.array(self._config.low)
        else:
            return np.ones(self.dimension()) * -1

    def limit_high(self):
        if self._config.high:
            return np.array(self._config.high)
        else:
            return np.ones(self.dimension()) * 1

    def position(self, **kwargs) -> np.ndarray:
        return np.array(self._config.desired_position)

    def velocity(self, **kwargs) -> np.ndarray:
        return np.zeros(self.dimension())

    def acceleration(self, **kwargs):
        return np.zeros(self.dimension())

    def shuffle(self):
        random_pos = np.random.uniform(
            self.limit_low(), self.limit_high(), self.dimension()
        )
        self._config.desired_position = random_pos.tolist()

    def angle(self):
        if isinstance(self._config.angle, float):
            return self._config.angle
        if self._config.angle:
            return list(self._config.angle)



from dataclasses import dataclass
from typing import Optional
from omegaconf import OmegaConf

from mpscenes.obstacles.dynamic_obstacle import DynamicObstacle, DynamicGeometryConfig
from mpscenes.obstacles.urdf_obstacle import UrdfGeometryConfig, UrdfObstacle, UrdfObstacleConfig


@dataclass
class DynamicUrdfGeometryConfig(UrdfGeometryConfig, DynamicGeometryConfig):
    pass

@dataclass
class DynamicUrdfObstacleConfig(UrdfObstacleConfig):
    """Configuration dataclass for sphere obstacle.

    This configuration class holds information about the position, size
    and randomization of a dynamic spherical obstacle.

    Parameters:
    ------------

    geometry : DynamicUrdfGeometryConfig : Geometry of the obstacle
    low: DynamicUrdfGeometryConfig : Lower limit for randomization
    high: DynamicUrdfGeometryConfig : Upper limit for randomization
    """

    geometry: DynamicUrdfGeometryConfig
    low: Optional[DynamicUrdfGeometryConfig] = None
    high: Optional[DynamicUrdfGeometryConfig] = None


class DynamicUrdfObstacle(DynamicObstacle, UrdfObstacle):

    def __init__(self, **kwargs):
        schema = OmegaConf.structured(DynamicUrdfObstacleConfig)
        kwargs['schema'] = schema
        super().__init__(**kwargs)



from typing import List, Optional
from dataclasses import dataclass, field
from abc import abstractmethod
import numpy as np

from mpscenes.common.component import MPComponent


def quaternion_to_homogeneous_matrix(
    translation: np.ndarray, quaternion: np.ndarray
) -> np.ndarray:
    w, x, y, z = quaternion

    # Normalize the quaternion
    norm = np.sqrt(w**2 + x**2 + y**2 + z**2)
    w /= norm
    x /= norm
    y /= norm
    z /= norm

    # Calculate transformation matrix elements
    xx = 1 - 2 * y**2 - 2 * z**2
    xy = 2 * x * y - 2 * w * z
    xz = 2 * x * z + 2 * w * y
    yx = 2 * x * y + 2 * w * z
    yy = 1 - 2 * x**2 - 2 * z**2
    yz = 2 * y * z - 2 * w * x
    zx = 2 * x * z - 2 * w * y
    zy = 2 * y * z + 2 * w * x
    zz = 1 - 2 * x**2 - 2 * y**2

    # Construct the transformation matrix
    transformation_matrix = np.array(
        [
            [xx, xy, xz, translation[0]],
            [yx, yy, yz, translation[1]],
            [zx, zy, zz, translation[2]],
            [0, 0, 0, 1],
        ]
    )

    return transformation_matrix


@dataclass
class GeometryConfig:
    """Configuration dataclass for geometry.

    This configuration class holds information about position and orientation
    of an obstacle. This class is further specified for the other obstacles.

    Parameters:
    ------------

    position: list: Position of the obstacle
    orientation: list: Orientation of the obstacle
    color: list: RGB-A color code
    """

    position: List[float] = field(default_factory=lambda: [0.0, 0.0, 0.0])
    orientation: List[float] = field(default_factory=lambda: [1.0, 0.0, 0.0, 0.0])


@dataclass
class CollisionObstacleConfig:
    """Configuration dataclass for sphere obstacle.

    This configuration class holds information about the dimension and the type
    of collision obstacle.

    Parameters:
    ------------

    type : str : Type of the obstacle
    geometry : GeometryConfig : Geometry of the obstacle
    movable : bool : Flag indicating whether an obstacle can be pushed around
    low : GeometryConfig : Lower limit for randomization
    high : GeometryConfig : Upper limit for randomization
    rgba: list: Color in rgba encoding
    """

    type: str
    geometry: GeometryConfig
    movable: bool = False
    low: Optional[GeometryConfig] = None
    high: Optional[GeometryConfig] = None
    rgba: List[float] = field(default_factory=lambda: [0.0, 0.0, 0.0, 1.0])


class CollisionObstacle(MPComponent):
    def type(self) -> str:
        return self._config.type

    def geometry(self) -> GeometryConfig:
        return self._config.geometry

    def evaluate(self, **kwargs) -> list:
        return [
            self.position(**kwargs),
            self.velocity(**kwargs),
            self.acceleration(**kwargs),
        ]

    def dimension(self) -> int:
        return len(self.geometry().position)

    def orientation(self, **kwarg) -> np.ndarray:
        return np.array(self.geometry().orientation)

    def position(self, **kwargs) -> np.ndarray:
        return np.array(self.geometry().position)

    def rgba(self) -> np.ndarray:
        return np.array(self._config.rgba)

    def velocity(self, **kwargs) -> np.ndarray:
        return np.zeros(self.dimension())

    def acceleration(self, **kwargs) -> np.ndarray:
        return np.zeros(self.dimension())

    def movable(self) -> bool:
        return self._config.movable

    def position_into_obstacle_frame(self, positions: np.ndarray, **kwargs) -> np.ndarray:
        if 't' in kwargs:
            t = kwargs.get('t')
        else:
            t = 0
        transformation_matrix = quaternion_to_homogeneous_matrix(
            self.position(t=t), self.orientation()
        )
        if len(positions.shape) > 1:
            nb_points = positions.shape[0]
            positions_homo = np.transpose(np.append(positions, np.ones((nb_points, 1)), axis=1))
            return np.dot(np.linalg.inv(transformation_matrix), positions_homo)[0:3,:]
        else:
            positions_homo = np.transpose(np.append(positions, 1))
            return np.dot(np.linalg.inv(transformation_matrix), positions_homo)[0:3]


    @abstractmethod
    def distance(self, position: np.ndarray) -> float:
        pass

    @abstractmethod
    def size(self) -> np.ndarray:
        pass



from mpscenes.goals.static_sub_goal import StaticSubGoal
from mpscenes.goals.static_joint_space_sub_goal import StaticJointSpaceSubGoal
from mpscenes.goals.dynamic_sub_goal import DynamicSubGoal
from mpscenes.common.errors import UnknownSubGoalType


class SubGoalCreator(object):
    def __init__(self):
        pass

    def create_sub_goal(self, sub_goal_type, name, content_dict):
        if sub_goal_type == "staticSubGoal":
            return StaticSubGoal(name=name, content_dict=content_dict)
        elif sub_goal_type in ("analyticSubGoal", "splineSubGoal"):
            return DynamicSubGoal(name=name, content_dict=content_dict)
        elif sub_goal_type == "staticJointSpaceSubGoal":
            return StaticJointSpaceSubGoal(name=name, content_dict=content_dict)
        else:
            raise UnknownSubGoalType(
                f"SubGoalType {sub_goal_type} is not known"
            )



from dataclasses import dataclass
from typing import List, Optional
import os
import csv
import numpy as np
from omegaconf import OmegaConf

from mpscenes.obstacles.collision_obstacle import CollisionObstacle, CollisionObstacleConfig, GeometryConfig


@dataclass
class SphereGeometryConfig(GeometryConfig):
    """Configuration dataclass for geometry.

    This configuration class holds information about position
    and radius of a sphere obstacle.

    Parameters:
    ------------

    radius: float: Radius of the obstacle
    """

    radius: float = 1.0


@dataclass
class SphereObstacleConfig(CollisionObstacleConfig):
    """Configuration dataclass for sphere obstacle.

    This configuration class holds information about the position, size
    and randomization of a spherical obstacle.

    Parameters:
    ------------

    geometry : GeometryConfig : Geometry of the obstacle
    low : GeometryConfig : Lower limit for randomization
    high : GeometryConfig : Upper limit for randomization
    """

    geometry: SphereGeometryConfig
    low: Optional[SphereGeometryConfig] = None
    high: Optional[SphereGeometryConfig] = None


class SphereObstacle(CollisionObstacle):
    def __init__(self, **kwargs):
        if 'schema' not in kwargs:
            schema = OmegaConf.structured(SphereObstacleConfig)
            kwargs['schema'] = schema
        super().__init__(**kwargs)
        self.check_completeness()

    def size(self):
        return [
            self.radius(),
        ]

    def limit_low(self):
        if self._config.low:
            return [
                np.array(self._config.low.position),
                self._config.low.radius,
            ]
        else:
            return [np.ones(self.dimension()) * -1, 0]

    def limit_high(self):
        if self._config.high:
            return [
                np.array(self._config.high.position),
                self._config.high.radius,
            ]
        else:
            return [np.ones(self.dimension()) * 1, 1]

    def radius(self):
        return self._config.geometry.radius

    def shuffle(self):
        random_pos = np.random.uniform(
            self.limit_low()[0], self.limit_high()[0], self.dimension()
        )
        random_radius = np.random.uniform(
            self.limit_low()[1], self.limit_high()[1], 1
        )
        self._config.geometry.position = random_pos.tolist()
        self._config.geometry.radius = float(random_radius)


    def csv(self, file_name, samples=100):
        theta = np.arange(-np.pi, np.pi + np.pi / samples, step=np.pi / samples)
        x = self.position()[0] + (self.radius() - 0.1) * np.cos(theta)
        y = self.position()[1] + (self.radius() - 0.1) * np.sin(theta)
        with open(file_name, "w", encoding="utf8") as file:
            csv_writer = csv.writer(file, delimiter=",")
            for i in range(2 * samples):
                csv_writer.writerow([x[i], y[i]])

    def distance(self, position: np.ndarray, **kwargs) -> float:
        pos = self.position_into_obstacle_frame(position, **kwargs)
        return np.linalg.norm(pos, axis=0) - self.radius()



from dataclasses import dataclass
from typing import Any

from mpscenes.obstacles.collision_obstacle import CollisionObstacle
from mpscenes.common.errors import MissmatchDimensionError, TrajectoryNotSupported
from mpscenes.common.analytic_trajectory import AnalyticTrajectory
from mpscenes.common.spline_trajectory import SplineTrajectory


@dataclass
class DynamicGeometryConfig:
    """Configuration dataclass for geometry.

    This configuration class holds information about position
    and radius of a dynamic sphere obstacle.

    Parameters:
    ------------

    trajectory: Any
        trajectory description of the obstacle. Can be either a spline or
        a analytic trajectory.
    radius: float
        radius of the obstacle
    """

    trajectory: Any


class DynamicObstacle(CollisionObstacle):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        if "controlPoints" in self.geometry()["trajectory"]:
            self._trajectory_type = "spline"
        elif isinstance(self.geometry()["trajectory"][0], str):
            self._trajectory_type = "analytic"
        else:
            raise TrajectoryNotSupported(
                f"Trajectory definition not supported, {self.geometry()['trajectory']}"
            )
        self.check_completeness()
        self.check_dimensionality()
        if self.trajectory_type() == "spline":
            self._traj = SplineTrajectory(
                self.dimension(), traj=self._config.geometry.trajectory
            )
        elif self.trajectory_type() == "analytic":
            self._traj = AnalyticTrajectory(
                self.dimension(), traj=self._config.geometry.trajectory
            )
        self._traj.concretize()

    def trajectory_type(self) -> str:
        return self._trajectory_type

    def check_dimensionality(self):
        dim_verification = self.dimension()
        if self.dimension() != dim_verification:
            raise MissmatchDimensionError(
                "Dynamic Obstacle: Dimension mismatch between trajectory array and dimension"
            )

    def dimension(self) -> int:
        if self.trajectory_type() == "spline":
            return len(self.geometry()["trajectory"]["controlPoints"][0])
        if self.trajectory_type() == "analytic":
            return len(self.geometry()["trajectory"])

    def traj(self):
        return self._traj

    def position(self, **kwargs):
        if "t" not in kwargs:
            t = 0.0
        else:
            t = kwargs.get("t")
        return self._traj.evaluate(t)[0]

    def velocity(self, **kwargs):
        if "t" not in kwargs:
            t = 0.0
        else:
            t = kwargs.get("t")
        return self._traj.evaluate(t)[1]

    def acceleration(self, **kwargs):
        if "t" not in kwargs:
            t = 0.0
        else:
            t = kwargs.get("t")
        return self._traj.evaluate(t)[2]
